#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "macros.h"
#include "list.h"
#include "as_tree.h"
#include "xmalloc.h"
#include "quad.h"
#include "symbol.h"
#include "list.h"

static void quad_release(struct quad *q)
{
	return_if_fail(q != NULL);
	xfree(q);
}

static void quad_free(struct quad *q)
{
	return_if_fail(q != NULL);

	if (q->release != NULL)
		q->release(q);
}

void quad_ref(struct quad *q)
{
	return_if_fail(q != NULL);

	++q->ref_count;
}

void quad_unref(struct quad *q)
{
	return_if_fail(q != NULL);
	return_if_fail(q->ref_count != 0);

	--q->ref_count;

	if (q->ref_count == 0)
		quad_free(q);
}

static struct quad *quad_new(size_t size)
{
	struct quad *q;

	q = xmalloc0(size);
	q->ref_count = 1;

	return q;
}

/* The destructor and the constructor of the constant argument */
static void quad_const_release(struct quad *q)
{
	struct quad_const *_const;

	return_if_fail(q != NULL);

	_const = (struct quad_const *)q;

	if (QUAD_ARG(_const)->data_type == DATA_TYPE_STRING)
		xfree(_const->str);

	xfree(_const);
}

struct quad_const *quad_const(data_type_t data_type, void *data)
{
	struct quad_const *_const;
	int *integer;
	float *real;
	char *ch, *str;

	_const = (struct quad_const *)quad_new(sizeof(*_const));

	QUAD(_const)->quad_type = QUAD_CONST;
	QUAD(_const)->release = quad_const_release;
	QUAD(_const)->ref_count = 1;
	QUAD_ARG(_const)->data_type = data_type;

	switch (data_type) {
	case DATA_TYPE_INTEGER:
		integer = data;
		_const->integer = *integer;
		break;
	case DATA_TYPE_REAL:
		real = data;
		_const->real = *real;
		break;
	case DATA_TYPE_CHAR:
		ch = data;
		_const->ch = *ch;
		break;
	case DATA_TYPE_STRING:
		str = data;
		_const->str = xstrdup(str);
		break;
	default:
		break;
	}

	return _const;
}

/* The destructor and the constructor of the id argument */
static void quad_id_release(struct quad *q)
{
	struct quad_id *id;

	return_if_fail(q != NULL);

	id = (struct quad_id *)q;

	xfree(id->id);
	xfree(id);
}

struct quad_id *quad_id(data_type_t decl_type, const char *name)
{
	struct quad_id *id;

	return_val_if_fail(name != NULL, NULL);

	id = (struct quad_id *)quad_new(sizeof(*id));
	QUAD(id)->quad_type = QUAD_ID;
	QUAD(id)->release = quad_id_release;
	QUAD(id)->ref_count = 1;

	QUAD_ARG(id)->data_type = decl_type;
	id->id = xstrdup(name);

	return id;
}

struct quad_refer *quad_refer(data_type_t decl_type, const char *name)
{
	struct quad_refer *refer;

	return_val_if_fail(name != NULL, NULL);

	refer = (struct quad_refer *)quad_new(sizeof(*refer));
	QUAD(refer)->quad_type = QUAD_REFER;
	QUAD(refer)->release = quad_id_release;
	QUAD(refer)->ref_count = 1;

	QUAD_ARG(refer)->data_type = decl_type;
	refer->id = xstrdup(name);

	return refer;
}

/* The destructor and the constructor of the if quad */
static void quad_if_release(struct quad *q)
{
	struct quad_if *if_quad;

	return_if_fail(q != NULL);

	if_quad = (struct quad_if *)q;

	quad_unref(if_quad->left);
	quad_unref(if_quad->right);
	xfree(if_quad);
}

struct quad_if *quad_if(opcode_t opcode, struct quad *left, struct quad *right, int ln)
{
	struct quad_if *if_quad;

	if_quad = (struct quad_if *)quad_new(sizeof(*if_quad));

	if_quad->opcode = opcode;
	if_quad->left = left;
	if_quad->right = right;
	if_quad->ln = ln;

	QUAD(if_quad)->quad_type = QUAD_IF;
	QUAD(if_quad)->release = quad_if_release;

	return if_quad;
}

/* A constructor of a goto quad, it uses common destructor */
struct quad_goto *quad_goto(int ln)
{
	struct quad_goto *goto_quad;

	goto_quad = (struct quad_goto *)quad_new(sizeof(*goto_quad));

	goto_quad->ln = ln;

	QUAD(goto_quad)->quad_type = QUAD_GOTO;
	QUAD(goto_quad)->release = quad_release;

	return goto_quad;
}

/* The constructor of the label quad, it uses a common destructor */
struct quad_label *quad_label(int ln)
{
	struct quad_label *label_quad;

	label_quad = (struct quad_label *)quad_new(sizeof(*label_quad));

	label_quad->ln = ln;

	QUAD(label_quad)->quad_type = QUAD_LABEL;
	QUAD(label_quad)->release = quad_release;

	return label_quad;
}

/* The destructor and the constructor of the function call quad */
static void quad_call_release(struct quad *q)
{
	struct quad_call *call_quad;

	return_if_fail(q != NULL);

	call_quad = (struct quad_call *)q;

	xfree(call_quad->name);
	if (call_quad->ret_val != NULL)
		quad_unref(call_quad->ret_val);

	if (call_quad->args != NULL)
		xfree(call_quad->args);

	xfree(call_quad);
}

struct quad_call *quad_call(const char *name, struct quad *ret_val, data_type_t ret_type)
{
	struct quad_call *call_quad;

	call_quad = (struct quad_call *)quad_new(sizeof(*call_quad));

	QUAD(call_quad)->quad_type = QUAD_CALL;
	QUAD(call_quad)->release = quad_call_release;

	call_quad->name = xstrdup(name);
	call_quad->ret_val = ret_val;
	call_quad->ret_type = ret_type;

	return call_quad;
}

void quad_call_add_arg(struct quad_call *call_quad, data_type_t arg_type, boolean_t is_ref)
{
	struct q_call_arg arg;

	arg.is_ref = is_ref;
	arg.arg_type = arg_type;

	call_quad->nargs++;
	call_quad->args = xrealloc(call_quad->args, sizeof(struct q_call_arg)*call_quad->nargs);
	call_quad->args[call_quad->nargs-1] = arg;
}

/* The destructor and the constructor of the function parameter quad */
static void quad_param_release(struct quad *q)
{
	struct quad_param *param_quad;;

	return_if_fail(q != NULL);

	param_quad = (struct quad_param *)q;
	quad_unref(param_quad->param);

	xfree(param_quad);
}

struct quad_param *quad_param(struct quad *param, boolean_t is_ref)
{
	struct quad_param *param_quad;

	return_val_if_fail(param != NULL, NULL);

	param_quad = (struct quad_param *)quad_new(sizeof(*param_quad));

	param_quad->param = param;
	param_quad->is_ref = is_ref;

	QUAD(param_quad)->quad_type = QUAD_PARAM;
	QUAD(param_quad)->release = quad_param_release;

	return param_quad;
}

/* The destructor and the constructor of the operation quad */
static void quad_op_release(struct quad *q)
{
	struct quad_op *op_quad;;

	return_if_fail(q != NULL);

	op_quad = (struct quad_op *)q;

	quad_unref(op_quad->result);
	quad_unref(op_quad->left);
	quad_unref(op_quad->right);

	xfree(op_quad);
}

struct quad_op *quad_op(opcode_t opcode, struct quad *result, struct quad *left, struct quad *right)
{
	struct quad_op *op_quad;

	return_val_if_fail(result != NULL, NULL);
	return_val_if_fail(left != NULL, NULL);
	return_val_if_fail(right != NULL, NULL);

	op_quad = (struct quad_op *)quad_new(sizeof(*op_quad));

	op_quad->opcode = opcode;
	op_quad->result = result;
	op_quad->left = left;
	op_quad->right = right;

	QUAD(op_quad)->quad_type = QUAD_OP;
	QUAD(op_quad)->release = quad_op_release;

	return op_quad;
}

/* The destructor and the constructor of the function quad */
static void quad_func_release(struct quad *q)
{
	struct quad_func *func_quad;

	return_if_fail(q != NULL);

	func_quad = (struct quad_func *)q;

	xfree(func_quad->name);
	xfree(func_quad);
}

struct quad_func *quad_func(const char *name, data_type_t ret_type)
{
	struct quad_func *func_quad;

	return_val_if_fail(name != NULL, NULL);

	func_quad = (struct quad_func *)quad_new(sizeof(*func_quad));

	func_quad->name = xstrdup(name);
	func_quad->ret_type = ret_type;

	QUAD(func_quad)->quad_type = QUAD_FUNC;
	QUAD(func_quad)->release = quad_func_release;

	return func_quad;
}

/* The destructor and the constructor of the assign quad */
static void quad_assign_release(struct quad *q)
{
	struct quad_assign *assign_quad;

	return_if_fail(q != NULL);

	assign_quad = (struct quad_assign *)q;

	quad_unref(assign_quad->lvalue);
	quad_unref(assign_quad->rvalue);

	xfree(assign_quad);
}

struct quad_assign *quad_assign(struct quad *lvalue, struct quad *rvalue)
{
	struct quad_assign *assign_quad;

	assign_quad = (struct quad_assign *)quad_new(sizeof(*assign_quad));

	assign_quad->lvalue = lvalue;
	assign_quad->rvalue = rvalue;

	QUAD(assign_quad)->quad_type = QUAD_ASSIGN;
	QUAD(assign_quad)->release = quad_assign_release;

	return assign_quad;
}

/* The destructor and the constructor of the store access quad */
static void quad_store_access_release(struct quad *q)
{
	struct quad_store_access *store;

	return_if_fail(q != NULL);

	store = (struct quad_store_access *)q;

	xfree(store->name);
	quad_unref(store->idx);
	quad_unref(store->rvalue);

	xfree(store);
}

struct quad_store_access *quad_store_access(const char *name, struct quad *idx, struct quad *rvalue)
{
	struct quad_store_access *store;

	return_val_if_fail(idx != NULL, NULL);
	return_val_if_fail(rvalue != NULL, NULL);

	store = (struct quad_store_access *)quad_new(sizeof(*store));

	store->name = xstrdup(name);
	store->idx = idx;
	store->rvalue = rvalue;

	QUAD(store)->quad_type = QUAD_STORE_ACCESS;
	QUAD(store)->release = quad_store_access_release;

	return store;
}

/* The destructor and the constructor of the load access quad */
static void quad_load_access_release(struct quad *q)
{
	struct quad_load_access *load;

	return_if_fail(q != NULL);

	load = (struct quad_load_access *)q;

	xfree(load->name);
	quad_unref(load->idx);
	quad_unref(load->lvalue);

	xfree(load);
}

struct quad_load_access	*quad_load_access(const char *name, struct quad *lvalue, struct quad *idx)
{
	struct quad_load_access *load;

	load = (struct quad_load_access *)quad_new(sizeof(*load));

	QUAD(load)->quad_type = QUAD_LOAD_ACCESS;
	QUAD(load)->release = quad_load_access_release;

	load->name = xstrdup(name);
	load->lvalue = lvalue;
	load->idx = idx;

	return load;
}

#ifdef _MEM_AREA
/* The destructor and the constructor of the store memory area quad */
static void quad_store_mem_area_release(struct quad *q)
{
	struct quad_store_mem_area *store;

	return_if_fail(q != NULL);

	store = (struct quad_store_mem_area *)q;

	quad_unref(store->rvalue);

	if (store->idx != NULL)
		quad_unref(store->idx);

	xfree(store);
}

struct quad_store_mem_area *quad_store_mem_area()
{
	struct quad_store_mem_area *store;

	store = (struct quad_store_mem_area *)quad_new(sizeof(*store));

	QUAD(store)->quad_type = QUAD_STORE_MEM_AREA;
	QUAD(store)->release = quad_store_mem_area_release;

	return store;
}

/* The destructor and the constructor of the load memory area quad */
static void quad_load_mem_area_release(struct quad *q)
{
	struct quad_load_mem_area *load;

	return_if_fail(q != NULL);

	load = (struct quad_load_mem_area *)q;

	quad_unref(load->lvalue);

	if (load->idx != NULL)
		quad_unref(load->idx);

	xfree(load);
}

struct quad_load_mem_area *quad_load_mem_area()
{
	struct quad_load_mem_area *load;

	load = (struct quad_load_mem_area *)quad_new(sizeof(*load));

	QUAD(load)->quad_type = QUAD_LOAD_MEM_AREA;
	QUAD(load)->release = quad_load_mem_area_release;

	return load;
}
#endif

static void quad_scope_release(struct quad *q)
{
	struct quad_scope *scope_quad;

	return_if_fail(q != NULL);

	scope_quad = (struct quad_scope *)q;
	symbol_table_unref(scope_quad->table);
	xfree(scope_quad);
}

struct quad_scope *quad_scope(struct symbol_table *table)
{
	struct quad_scope *scope_quad;

	return_val_if_fail(table != NULL, NULL);

	scope_quad = (struct quad_scope *)quad_new(sizeof(*scope_quad));

	QUAD(scope_quad)->quad_type = QUAD_SCOPE;
	QUAD(scope_quad)->release = quad_scope_release;
	scope_quad->table = table;

	return scope_quad;
}

struct quad_scope_end *quad_scope_end()
{
	struct quad_scope_end *scope_end_quad;

	scope_end_quad = (struct quad_scope_end *)quad_new(sizeof(*scope_end_quad));

	QUAD(scope_end_quad)->quad_type = QUAD_SCOPE_END;
	QUAD(scope_end_quad)->release = quad_release;

	return scope_end_quad;
}

struct quad_func_end *quad_func_end()
{
	struct quad_func_end *func_end_quad;

	func_end_quad = (struct quad_func_end *)quad_new(sizeof(*func_end_quad));

	QUAD(func_end_quad)->quad_type = QUAD_FUNC_END;
	QUAD(func_end_quad)->release = quad_release;

	return func_end_quad;
}

